/*
 *  bf-crypt.cpp
 *
 *  BLOWFISH CRYPTOGRAPHY
 *
 *  Copyright (c) 2006 by Dominik Erdmann - www.der-domi.de
 *
 *  Description:
 *  bf-crypt.exe encrypt/decrpyt files by using the blowfish algorithm.
 *  The encrypted file will be base64-converted.
 *
 *  Used projects:
 *  Your password will be hashed with RipeMD-160 
 *  from http://homes.esat.kuleuven.be/~bosselae/ripemd160.html.
 *  So the keylenght is always 160 bits.
 *  For blowfish cryptography I used the reference source code from SSLeay 0.6.6
 *  For base64 I used the source code from http://base64.sourceforge.net
 *
 *  encrypt - syntax:
 *  bf-crypt -e <file-to-encrypt> <encrypted-file>
 *
 *  decrypt - syntax:
 *  bf-crypt -d <file-to-decrypt> <decrypted-file>
 *
 */


/*
 *  includes
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>		// for getch(): this does not work with linux. sorry!

#include "bf-crypt.h"
#include "blowfish.h"
#include "b64.h"
#include "rmd160.h"


/*
 *  main
 */

int main(int argc, char* argv[])
{
	unsigned char *buffer=NULL, *cipherBuffer=NULL;
	int mode, lenBuffer, lenCipher, ret;
	BLOWFISH_CTX ctx;

	printInfo();

	if ( argc != 4 )
	{
		printf("\nERROR: Wrong count of parameter\n");
		exit(-1);
	}
	if ( argv[1][0] == '-' && argv[1][1] == 'e' )
		mode = ENCRYPT;
	else if ( argv[1][0] == '-' && argv[1][1] == 'd' )
		mode = DECRYPT;
	else
	{
		printf("\nERROR: Wrong mode-parameter\n");
		exit(-1);
	}


	switch ( mode )
	{
		case ENCRYPT:
		{
			buffer = fillBufferFromFile( argv[2], &lenBuffer );
			cipherBuffer = encodeBuffer( buffer, lenBuffer, &lenCipher );
			clearBuffer( buffer, lenBuffer );
			ctx_init( &ctx, mode );
			crypt( &ctx, cipherBuffer, lenCipher, mode );
			fillFileFromBuffer( cipherBuffer, lenCipher, TMP_FILENAME );
			clearBuffer( cipherBuffer,lenCipher );
			ret = b64( 'e', TMP_FILENAME, argv[3], 70 );

			if ( remove( TMP_FILENAME ) )
			{
				printf("\nERROR: Can't delete temporary file '%s'\n", TMP_FILENAME);
				exit(-1);
			}

			printf("\n\nYour File is encrypted now.\n");

			break;
		}

		case DECRYPT:
		{
			ret = b64( 'd', argv[2], TMP_FILENAME, 70 );
			cipherBuffer = fillBufferFromFile( TMP_FILENAME, &lenCipher );
			if ( remove( TMP_FILENAME ) )
			{
				printf("\nERROR: Can't delete temporary file '%s'\n", TMP_FILENAME);
				exit(-1);
			}
			ctx_init( &ctx, mode );
			crypt( &ctx, cipherBuffer, lenCipher, mode );
			buffer = decodeBuffer( cipherBuffer, lenCipher, &lenBuffer );
			clearBuffer( cipherBuffer, lenCipher );
			fillFileFromBuffer( buffer, lenBuffer, argv[3] );
			clearBuffer( buffer, lenBuffer );

			printf("\n\nYour File is decrypted now.\n");
			break;
		}
	}

	return 0;
}


/*
 *  functions
 */


/*
 *  expands a given length to next 64 bit
 */
int getNext64BitModulo( int lenght ) 
{
	int mod;
	mod = lenght % (64/8);

	if ( mod == 0 )
		return lenght;
	else
		return lenght + 8 - mod;
}


/*
 *  get password from stdin
 */
int getPwdEntry( char *pwd )
{
	int i = 0;
	char tmp;

	do {
		tmp = getch(); // waiting for input

		if ( tmp == ABORT )
			exit(-1);
		if ( tmp != LINEFEED && i < MAX_PWD_LEN ) {
			pwd[i] = tmp;
			i++;
		}
		if ( tmp == BACKSPACE ) {
			i = i - 2;
			if ( i < 0 )
				i = 0;
			pwd[i] = NUL;
		}
		if ( tmp == LINEFEED && i < MIN_PWD_LEN ) {
			printf("%c", BEEP);
		}
		if ( tmp != LINEFEED && i >= MAX_PWD_LEN ) {
			printf("%c", BEEP);
		}
	} while ( tmp != LINEFEED || i < MIN_PWD_LEN && i < MAX_PWD_LEN );

	// terminate string
	pwd[i] = NUL;

	return i;
}


/*
 *  check password against mode and entry
 */
int getPwd( char *pwd, int mode )
{
	int lenPwd;
	char tmpPwd[MAX_PWD_LEN+1];

	while (1)
	{
		printf("\nEnter your password: ");
		lenPwd = getPwdEntry(pwd);
		if ( mode == DECRYPT )
			return lenPwd;
		else
		{
			printf("\nRepeat password:     ");
			getPwdEntry(tmpPwd);
			if ( strcmp( pwd, tmpPwd ) == 0 )
			{
				clearBuffer( (unsigned char*)tmpPwd, MAX_PWD_LEN );
				return lenPwd;
			}
			else
				printf("\n\nUnequal entry. Try again. CRTL-C to exit.\n");
		}
		clearBuffer( (unsigned char*)tmpPwd, MAX_PWD_LEN );
	}
}


/*
 *  read file and return size in bytes
 */
int getFileSize( FILE *file )
{
	int size;
	fseek (file , 0 , SEEK_END);
	size = ftell (file);
	rewind (file);
	return size;
}


/*
 *  read file and write characters in buffer
 */
unsigned char *fillBufferFromFile( char *filename, int *lenBuffer )
{
	int fileSize;
	FILE *file;
	unsigned char *buffer;

	if ( (file = fopen( filename, "rb" )) == NULL )
	{
		printf("\nERROR: Can't open file '%s' to read\n", filename);
		exit(-1);
	}

	fileSize = getFileSize(file);
	buffer = (unsigned char*)malloc(fileSize);
	*lenBuffer = (int)fread( buffer, 1, fileSize, file );
	fclose(file);

	if ( *lenBuffer != fileSize )
	{
		printf("\nERROR: Problem with reading file '%s'\n", filename );
		exit(-1);
	}
	return buffer;
}


/*
 *  read buffer and write characters in file
 */
void fillFileFromBuffer( unsigned char *buffer, int lenBuffer, char *filename )
{
	int ret;
	FILE *file;

	if ( (file = fopen( filename, "wb" )) == NULL )
	{
		printf("\nERROR: Can't open file '%s' to write\n", filename);
		exit(-1);
	}

	ret = (int)fwrite( buffer, 1, lenBuffer, file );
	fclose(file);
	if ( ret != lenBuffer )
	{
		printf("\nERROR: Problem with writing file '%s'\n", filename );
		exit(-1);
	}
}


/*
 *  expands a given buffer to next 64 bit for encrypt with blowfish
 */
unsigned char *encodeBuffer( unsigned char *buffer, int lenOld, int *lenNew )
{
	int i;
	unsigned char *newBuffer;

	*lenNew = getNext64BitModulo(lenOld);
	newBuffer = (unsigned char*)malloc(*lenNew);

	// copy plaintext
	for ( i = 0; i < lenOld; i++ )
		newBuffer[i] = buffer[i];

	// fill empty bytes with dummy
	for ( ; i < *lenNew; ++i )
		newBuffer[i] = 0x00;

	return newBuffer;
}


/*
 *  delete the expansion from decrypt
 */
unsigned char *decodeBuffer( unsigned char *buffer, int lenOld, int *lenNew )
{
	int i;
	unsigned char *newBuffer;

	for ( i = lenOld; buffer[i-1] == 0x00; i-- )
		;

	*lenNew = i;
	newBuffer = (unsigned char*)malloc(*lenNew);

	// copy ciphertext
	for ( i=0; i<*lenNew; i++ )
		newBuffer[i] = buffer[i];

	return newBuffer;
}


/*
 *  encrypt/decrypt a given buffer
 */
void crypt( BLOWFISH_CTX *ctx, unsigned char *buffer, int len, int mode )
{
	int i, j;
	unsigned char L[4], R[4];

	for ( i=0; i<len; i=i+8 )
	{
		for ( j=0; j<4; j++ )
		{
			L[j] = buffer[i+j];
			R[j] = buffer[i+j+4];
		}

		if ( mode == ENCRYPT )
			Blowfish_Encrypt( ctx, (unsigned long*)&L, (unsigned long*)&R );
		else
			Blowfish_Decrypt( ctx, (unsigned long*)&L, (unsigned long*)&R );

		for ( j=0; j<4; j++ )
		{
			buffer[i+j]   = L[j];
			buffer[i+j+4] = R[j];
		}
	}
}


/*
 *  clear a secret buffer
 */
void clearBuffer( unsigned char *buffer, int len )
{
	for ( int i = 0; i < len; i++ )
		buffer[i] = NUL;
}


/*
 *  print info
 */
void printInfo(void)
{
	printf("\nThis Programm encrypt or decrypt files by using the blowfish algorithm.");
	printf("\nCopyright 2006 - Dominik Erdmann - www.der-domi.de");
	printf("\nUsage: bf-crypt <-e|-d> <input-file> <output-file>");
	printf("\nYour passwort have to be equal or more then %d characters.", MIN_PWD_LEN);
	printf("\n");
}


/*
 * returns RMD(message)
 */
unsigned char *RMD(byte *message)
{
	dword         MDbuf[RMDsize/32];   /* contains (A, B, C, D(, E))   */
	static byte   hashcode[RMDsize/8]; /* for final hash-value         */
	dword         X[16];               /* current 16-word chunk        */
	unsigned int  i;                   /* counter                      */
	dword         length;              /* length in bytes of message   */
	dword         nbytes;              /* # of bytes not yet processed */

	/* initialize */
	MDinit(MDbuf);
	length = (dword)strlen((char *)message);

	/* process message in 16-word chunks */
	for (nbytes=length; nbytes > 63; nbytes-=64) {
		for (i=0; i<16; i++) {
			X[i] = BYTES_TO_DWORD(message);
			message += 4;
		}
		compress(MDbuf, X);
	}                                    /* length mod 64 bytes left */

	/* finish: */
	MDfinish(MDbuf, message, length, 0);

	for (i=0; i<RMDsize/8; i+=4) {
		hashcode[i]   = (byte) MDbuf[i>>2];         /* implicit cast to byte  */
		hashcode[i+1] = (byte)(MDbuf[i>>2] >>  8);  /*  extracts the 8 least  */
		hashcode[i+2] = (byte)(MDbuf[i>>2] >> 16);  /*  significant bits.     */
		hashcode[i+3] = (byte)(MDbuf[i>>2] >> 24);
	}

	return (unsigned char *)hashcode;
}


void ctx_init( BLOWFISH_CTX *ctx, int mode )
{
	char pwd[MAX_PWD_LEN+1];
	unsigned char *hash;
	int len;

	hash = (unsigned char*)malloc(RMDsize/8);

	len = getPwd( pwd, mode );
	hash = RMD( (byte*)pwd );
	clearBuffer( (unsigned char*)pwd, (int)strlen(pwd) );
	Blowfish_Init( ctx,hash, RMDsize/8 );
	clearBuffer( hash, RMDsize/8 );
}
